/* PBM Reader - reads PBM files and displays them on the screen.
   Version 3.3
   TODO read files smaller than Joyce resolution
*/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <cpm.h>
#include "input2.h"
#include "box.h"
#include "kernel.h"
#include "screen.h"
#include "pbmin.h"
#include "buffered_io.h"
#include "mytypes.h"

#define SEARCH_FIRST 0x11
#define SEARCH_NEXT  0x12
#define SET_DMA      0x1A
#define MAX_FILES    100
*/

unsigned char FCB[27] @ 0x0060;
unsigned char DMA[267];
char drive[3];
char fileName[15];
unsigned char diashowDelay;
errorCode_t erroCode;
bool invert;
char ch;

/* Copies a filename from the DMA array to the FileName variable.
 */
void dmaToFileName(int start) {
    int loop;
    char ch;
    char *cp;
    
    cp = fileName;
    *cp = '\0';
    for (loop = start + 1; loop < start + 9; loop++) { // copy filename
        ch = DMA[loop];
        if (ch != ' ') {
            *cp = ch;
            ++cp;
        }
    }
    *cp = '.';
    ++cp;
    for (loop = start + 9; loop < start + 12; loop++) { // copy extension
        ch = DMA[loop];
        if (ch != ' ') {
            *cp = ch;
            ++cp;
        }
    }
    *cp = '\0';
}

/* Reads VEC filenames into the fileNames array.
   Returns the number of files found.
 */
int readDirectory(char **fileNames) {
    int error, loop, start;
    int idx;
    
//  error := BDos(SetDMA, Addr(DMA)); // TODO
    error = bdos(CPM_SDMA, DMA);
    
    FCB[0] = 0;
    // filter: files with extension mask PBM only
    for (loop = 1; loop < 9; loop++) {
        FCB[loop] = '?';
    }
    FCB[ 9] = 'P';
    FCB[10] = 'B';
    FCB[11] = 'M';

    idx = 1;
//  error:= BDos(SearchFirst, Addr(FCB)); // TODO
    error = bdos(CPM_FFST, FCB);
    if (error != 255) {
        start = error * 32;
        dmaToFileName(start);
        strcpy(fileNames[idx], fileName);
        ++idx;
    }
    do {
//      error := BDos(SearchNext); // TODO
        error = bdos(CPM_FNXT, 0);
        start = error * 32;
        if (error != 255) {
            dmaToFileName(start);
            strcpy(fileNames[idx], fileName);
            ++idx;
            if (idx > MAX_FILES) {
              error = 255; // exit loop
            }
        }
    } while (error != 255);

    return --idx;
}


/* Display diashow.
 */
void diashow(void) {
    bool run;
    int i, j;
    char drive[3];
    char fileNames[MAX_FILES][15];
    int fileCount;
    char ch;
    char tmpFileName[15];

    gx_init();
    gx_hideCursor();

    fileCount = readDirectory(fileNames);
    if (fileCount > MAX_FILES) {
        fileCount = MAX_FILES;
    }
    run = true;

    while (run) {
        for (i = 0; i < fileCount; i++) {
            strcpy(fileName, fileNames[i]);
            if (strlen(drive) == 2) {
                strcpy(tmpFileName, fileName);
                strcpy(fileName, drive);
                strcat(fileName, tmpFileName);
            }
            loadScreenFromPBM(fileName, invert);
            
            for (j = 0; j < diashowDelay; j++) {
                msleep(1000);
                if (kbhit()) {
                    goto ExitDiashow;
                }
            }
        }
    }
ExitDiashow:
}


/* Inverts the screen in memory, i.e. without reading the file again.
 */
void invertInMemory(void) {
    unsigned char row;
    int i;

    for (row = 0; row < 32; row++) {
        gx_getScreenRow(row);
        for (i = 0; i < 720; i++) {
            gx_buffer[i] = gx_buffer[i] ^ 0xFF; // XOR
        }
        gx_putScreenRow(row);
    }
}


void clearCharLeft(void) {
    putchar(8);
    putchar(' ');
    putchar(8);
}

/* Inputs an 8.3 filename with optional drive letter; input is finished with
   one of the following keys: Enter, Tab, Down, Up.
   Backspace can be used to correct the input.
   Parameter: fileNameBuffer  char array with at least 15 elements; will
                              contain the filename entered
 */
void inputFileName(char *fileNameBuffer) {
    char ch;
    int cntFilename;
    int cntExtension;
    bool dotEntered;
    bool colonEntered;
    int len;
    char *pBuf; // points to the end of fileNameBuffer, i.e. to the '\0' character

    cntFilename = 0;
    cntExtension = 0;
    dotEntered = false;
    colonEntered = false;
    *fileNameBuffer = '\0';
    pBuf = fileNameBuffer;
    
    do {
        ch = getchar();
        len = strlen(fileNameBuffer);
        switch (ch) {
            case 127: // backspace
                if (len > 0) {
                    switch (fileNameBuffer[len - 1]) {
                        case '.': dotEntered = false;
                                  break;
                        case ':': colonEntered = false;
                                  break;
                        default:
                            if (dotEntered) 
                                --cntExtension;
                            else
                                --cntFilename;
                    }
                    --pBuf;
                    *pBuf = '\0'; // delete last character
                    putchar(8);
                    putchar(8);
                    putchar(' ');
                    putchar(' ');
                    putchar(8);
                    putchar(8);
                }
                else {
                    putchar(' '); // undo cursor left
                }
                break;
            case 58: // ':'
                if (!colonEntered && len == 1) {
                    colonEntered = true;
                    *pBuf = ch;
                    ++pBuf;
                    *pBuf = '\0';
                }
                else {
                    clearCharLeft();
                }
                break;
            case 46: // '.'
                if (!dotEntered && (colonEntered && len > 2 ||
                                   !colonEntered && len > 0)) {
                    dotEntered = true;
                    *pBuf = ch;
                    ++pBuf;
                    *pBuf = '\0';
                }
                else {
                    clearCharLeft();
                }
                break;
            case 10: // Enter
                break;
            default:
                if (ch >= 'A' && ch <= 'Z' ||
                    ch >= 'a' && ch <= 'z' ||
                    ch >= '0' && ch <= '9' ||
                    ch == '_') { // allowed characters
                    if (dotEntered) {
                        if (cntExtension < 3) {
                            *pBuf = ch;
                            ++pBuf;
                            *pBuf = '\0';
                            ++cntExtension;
                        }
                        else {
                            clearCharLeft();
                        }
                    }
                    else { // dotEntered = false
                        if (cntFilename < 9) {
                            *pBuf = ch;
                            ++pBuf;
                            *pBuf = '\0';
                            ++cntFilename;
                        }
                        else {
                            clearCharLeft();
                        }
                    }
                }
                else { // not allowed
                    clearCharLeft();
                }
        }
    } while(ch != 10 && // Enter
            ch != 9  && // Tab
            ch != 30 && // Down
            ch != 31);  // Up
}


/* Main program.
 */
int main(int argc, char *argv[]) {
    errorCode err;
    
    clearScreen();
    invert = false;
    diashowDelay = 5;

    // get the name of the PBM file
    if (argc > 1) {
        if (strcmp(argv[1], "-DIASHOW") == 0)  {
            if (argc > 2) {
                if (strlen(argv[2]) == 2) {
                    strcpy(drive, argv[2]);
                    // check if drive is one of A: to Z:
                    if (*drive < 'A' || *drive > 'Z' || drive[1] != ':') {
                        *drive = '\0'; // drive spec wrong - use current drive
                    }
                }
            }
            else {
                *drive = '\0'; // no drive provided - use current drive
            }
            diashow();
            goto byebye;
        }
        else {
            strcpy(fileName, argv[1]); // first commandline parameter treated as filename
            if (argc > 2) {
                if (strcmp(argv[2], "-INVERT") == 0) invert = true;
            }
        }
    }
    else {
        drawBox(0, 0, 90, 3);
        gotoXY(27, 1);
        printf("PBM file reader V3.3 by Bernd Bock");
        drawBox(0, 3, 90, 3);
        gotoXY(2, 4);
        printf("Enter name of PBM file: ");
        inputFileName(fileName);
        gotoXY(0, 6);
        if (*fileName == '\0') goto byebye; // abort if no filename entered
        printf("Invert the picture? (Y/N) ");
        ch = getch();
        if (ch == 'Y' || ch == 'y') invert = true;
    }
  
    gx_init();
    gx_hideCursor();

    err = loadScreenFromPBM(fileName, invert);
    if (err != ErrOk) {
        putchar('\n');
        switch(err) {
            case ErrCouldNotOpenFile:   printf("Could not open file %s\n", fileName);
            case ErrFormatNotSupported: printf("Format not supported: only PBM type 4\n");
            default:                    printf("Unknown error!");
        }
        writeKeyMessage();
        waitForKey();
    }
    else {
        do {
            ch = getch();
            if (ch == 'i') ch = 'I';            // 'I' input always uppercase
            if (ch == 'I') invertInMemory();
        } while (ch == 'I');
    }
  
byebye:
    statusBarOff();      // switch off status bar
    clearScreen();       // now clears the status bar, too
    statusBarOn();       // switch on status bar
    showCursor();
}
